#!/usr/bin/env python
import numpy as np
from sklearn.linear_model import LinearRegression


def format_img(img):
    assert img.ndim == 3
    Y, X, C = img.shape
    Y -= Y % 2
    X -= X % 2
    assert C in (3, 4)
    C = 3
    img = img[:Y, :X, :C]
    if np.issubdtype(img.dtype, np.floating):
        pass
    elif np.issubdtype(img.dtype, np.uint8):
        img = img / 255
    elif np.issubdtype(img.dtype, np.uint16):
        img = img / (2**16 - 1)
    elif np.issubdtype(img.dtype, np.uint32):
        img = img / (2**32 - 1)
    elif np.issubdtype(img.dtype, np.uint64):
        img = img / (2**64 - 1)
    else:
        raise TypeError(img.dtype)
    return img


def mosaic(img, δx, δy):
    Y, X, C = img.shape
    mos = img[:, :, 1].copy()
    mos[δy::2, δx::2] = img[δy::2, δx::2, 0]
    mos[1 - δy::2, 1 - δx::2] = img[1 - δy::2, 1 - δx::2, 2]
    return mos


def get_residuals(img, n=3):
    Y, X, _ = img.shape
    Yo = Y // 2 - 2 * n
    Xo = X // 2 - 2 * n
    residuals = np.zeros((4, Yo, Xo, 3))
    for i_grid, (δy, δx) in enumerate([(0, 0), (1, 1), (0, 1), (1, 0)]):
        mos = mosaic(img, δx, δy)
        input_values = np.stack([
            np.roll(mos, (φx, φy), axis=(0, 1)) for φx in range(-n, n + 1)
            for φy in range(-n, n + 1)
        ])
        # shape (2n+1**2, Y, X)
        inp_r = input_values[:, δy::2,
                             δx::2][:, n:-n,
                                    n:-n].reshape(input_values.shape[0], -1).T
        inp_g0 = input_values[:, δy::2,
                              1 - δx::2][:, n:-n,
                                         n:-n].reshape(input_values.shape[0],
                                                       -1).T
        inp_g1 = input_values[:, 1 - δy::2,
                              δx::2][:, n:-n,
                                     n:-n].reshape(input_values.shape[0], -1).T
        inp_b = input_values[:, 1 - δy::2,
                             1 - δx::2][:, n:-n,
                                        n:-n].reshape(input_values.shape[0],
                                                      -1).T

        targ_r_g = img[δy::2, δx::2, 1][n:-n, n:-n].ravel()[None].T
        targ_r_b = img[δy::2, δx::2, 2][n:-n, n:-n].ravel()[None].T
        targ_g0_r = img[δy::2, 1 - δx::2, 0][n:-n, n:-n].ravel()[None].T
        targ_g0_b = img[δy::2, 1 - δx::2, 2][n:-n, n:-n].ravel()[None].T
        targ_g1_r = img[1 - δy::2, δx::2, 0][n:-n, n:-n].ravel()[None].T
        targ_g1_b = img[1 - δy::2, δx::2, 2][n:-n, n:-n].ravel()[None].T
        targ_b_r = img[1 - δy::2, 1 - δx::2, 0][n:-n, n:-n].ravel()[None].T
        targ_b_g = img[1 - δy::2, 1 - δx::2, 1][n:-n, n:-n].ravel()[None].T
        residuals[i_grid, :, :, 0] = (np.square(
            LinearRegression(fit_intercept=True).fit(inp_g0, targ_g0_r).
            predict(inp_g0) - targ_g0_r) + np.square(
                LinearRegression(fit_intercept=True).fit(inp_g1, targ_g1_r).
                predict(inp_g1) - targ_g1_r) + np.square(
                    LinearRegression(fit_intercept=True).fit(inp_b, targ_b_r).
                    predict(inp_b) - targ_b_r)).reshape(Yo, Xo) / 3
        residuals[i_grid, :, :, 1] = (np.square(
            LinearRegression(fit_intercept=True).fit(inp_r, targ_r_g).predict(
                inp_r) - targ_r_g) + np.square(
                    LinearRegression(fit_intercept=True).fit(inp_b, targ_b_g).
                    predict(inp_b) - targ_b_g)).reshape(Yo, Xo) / 2
        residuals[i_grid, :, :, 2] = (np.square(
            LinearRegression(fit_intercept=True).fit(inp_g0, targ_g0_b).
            predict(inp_g0) - targ_g0_r) + np.square(
                LinearRegression(fit_intercept=True).fit(inp_g1, targ_g1_b).
                predict(inp_g1) - targ_g1_b) + np.square(
                    LinearRegression(fit_intercept=True).fit(inp_r, targ_r_b).
                    predict(inp_r) - targ_r_b)).reshape(Yo, Xo) / 3
    return residuals


def get_grids(residuals):
    diag = np.argmin(residuals[:, :, :, 1], axis=0) // 2
    diag_is_equal_green = np.abs(
        np.min(residuals[:2, :, :, 1], axis=0) -
        np.min(residuals[2:, :, :, 1], axis=0)) < (1 / 255**2) / 2 / 2
    res_rb = np.sum(np.delete(residuals, 1, axis=-1), axis=-1)
    diag_all = np.argmin(res_rb, axis=0) // 2
    diag[diag_is_equal_green] = diag_all[diag_is_equal_green]
    diag_is_equal_all = np.abs(
        np.min(res_rb[:2], axis=0) -
        np.min(res_rb[2:], axis=0)) < (1 / 255**2) / 6 / 2
    res_rb = res_rb[:2] * (diag == 0) + res_rb[2:] * (diag == 1)
    best_rb = np.argmin(res_rb, axis=0)
    rb_is_equal = np.abs(res_rb[0] - res_rb[1]) < (1 / 255**2) / 6 / 2
    grid = 2 * diag + best_rb
    grid[rb_is_equal] = -2 - diag[rb_is_equal]
    grid[diag_is_equal_green * diag_is_equal_all] = -1
    return grid


if __name__ == "__main__":
    from sys import argv
    from argparse import ArgumentParser
    from imageio import imread
    parser = ArgumentParser()
    parser.add_argument('input', type=str)
    parser.add_argument('output', type=str)
    parser.add_argument('-n', type=int, default=3)
    args = parser.parse_args(argv[1:])
    img = imread(args.input)
    img = format_img(img)
    n = args.n
    residuals = get_residuals(img, n)
    grids = get_grids(residuals)
    out = args.output
    np.savez_compressed(out,
                        grids=grids,
                        border=n,
                        down=n**2,
                        target_shape=img.shape[:2])
