#!/usr/bin/env python3
from argparse import ArgumentParser
from sys import argv
import numpy as np
from imageio import imread, imsave
import seaborn as sns
from matplotlib import colors, pyplot as plt
from matplotlib.colors import ListedColormap
from mpl_toolkits.axes_grid1 import Size, Divider

from get_block_votes import format_img, get_residuals, get_grids
from nfa_from_blocks import nfa_from_block_votes, interpolate_nfa


def create_fig(figsize):
    """Creates a figure, axis and colorbar axis such that the main axis has the specified size after saving."""
    px = 1 / plt.rcParams['figure.dpi']
    sY, sX = figsize

    pad = sX * 0.05
    cbsz = sX * 0.05
    fig = plt.figure(figsize=((sX + pad + cbsz) * px, sY * px))

    # The first items are for padding and the second items are for the axes.
    # sizes are in inch.
    h = [Size.Fixed(0 * px), Size.Fixed(sX * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (0, 0, 1 - (pad + cbsz) / (sY + pad + cbsz), 1),
                      h,
                      v,
                      aspect=False)
    # The width and height of the rectangle are ignored.
    ax = fig.add_axes(divider.get_position(),
                      axes_locator=divider.new_locator(nx=1, ny=1))
    h = [Size.Fixed(0 * px), Size.Fixed(cbsz * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (1 - (pad + cbsz) / (sX + pad + cbsz), 0, 1, 1),
                      h,
                      v,
                      aspect=False)
    cax = fig.add_axes(divider.get_position(),
                       axes_locator=divider.new_locator(nx=1, ny=1))
    return fig, ax, cax


grid_cmap_colours = ['black', '#da8bc3', '#8c8c8c', '#8172b3', '#64b5cd']
cmap_grid = ListedColormap(sns.color_palette(grid_cmap_colours).as_hex())
cmap_diag = ListedColormap(sns.color_palette(grid_cmap_colours[:3]).as_hex())


def save_grid(out, grid, figsize):
    fig, ax, cax = create_fig(figsize)
    plot = ax.matshow(grid, cmap=cmap_grid, vmin=-1, vmax=3)
    cb = plt.colorbar(plot, cax=cax)
    cax.tick_params(labelsize=35)
    cb.set_ticks([-.6, .2, 1, 1.8, 2.6])
    cb.set_ticklabels(['/', 'RGGB', 'BGGR', 'GRBG', 'GBRG'])
    ax.axis('off')
    plt.savefig(out, bbox_inches='tight', pad_inches=0, transparent=False)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument('input', type=str)
    parser.add_argument('n', type=int)
    parser.add_argument('threshold', type=float)
    args = parser.parse_args(argv[1:])
    img = imread(args.input)
    Y, X, _ = img.shape
    img = format_img(img)
    n = args.n
    residuals = get_residuals(img, n)
    VMAX = residuals.max() * 255
    r_rggb, r_bggr, r_grbg, r_gbrg = residuals * 255
    grids = get_grids(residuals)
    save_grid('grids.png', grids, (Y, X))

    nfa_d, nfa_g = nfa_from_block_votes(grids, n**2)
    nfa = np.minimum(nfa_d, nfa_g)
    nfa = interpolate_nfa(nfa, n, (Y, X))
    nfa = nfa < args.threshold
    nfa = nfa.astype(np.uint8) * 255
    imsave('detection.png', nfa)
